@@ -5,7 +5,9 @@
 package testscript
 
 import (
+	"bytes"
 	"errors"
+	"flag"
 	"fmt"
 	"io/ioutil"
 	"os"
@@ -25,7 +27,7 @@ func printArgs() int {
 	return 0
 }
 
-func echo() int {
+func fprintArgs() int {
 	s := strings.Join(os.Args[2:], " ")
 	switch os.Args[1] {
 	case "stdout":
@@ -47,7 +49,7 @@ func signalCatcher() int {
 	signal.Notify(c, os.Interrupt)
 	// Create a file so that the test can know that
 	// we will catch the signal.
-	if err := ioutil.WriteFile("catchsignal", nil, 0666); err != nil {
+	if err := ioutil.WriteFile("catchsignal", nil, 0o666); err != nil {
 		fmt.Println(err)
 		return 1
 	}
@@ -59,7 +61,7 @@ func signalCatcher() int {
 func TestMain(m *testing.M) {
 	os.Exit(RunMain(m, map[string]func() int{
 		"printargs":     printArgs,
-		"echo":          echo,
+		"fprintargs":    fprintArgs,
 		"status":        exitWithStatus,
 		"signalcatcher": signalCatcher,
 	}))
@@ -75,7 +77,7 @@ func TestCRLFInput(t *testing.T) {
 	}()
 	tf := filepath.Join(td, "script.txt")
 	contents := []byte("exists output.txt\r\n-- output.txt --\r\noutput contents")
-	if err := ioutil.WriteFile(tf, contents, 0644); err != nil {
+	if err := ioutil.WriteFile(tf, contents, 0o644); err != nil {
 		t.Fatalf("failed to write to %v: %v", tf, err)
 	}
 	t.Run("_", func(t *testing.T) {
@@ -132,7 +134,8 @@ func TestScripts(t *testing.T) {
 	// TODO set temp directory.
 	testDeferCount := 0
 	Run(t, Params{
-		Dir: "testdata",
+		UpdateScripts: os.Getenv("TESTSCRIPT_UPDATE") != "",
+		Dir:           "testdata",
 		Cmds: map[string]func(ts *TestScript, neg bool, args []string){
 			"setSpecialVal":    setSpecialVal,
 			"ensureSpecialVal": ensureSpecialVal,
@@ -175,12 +178,20 @@ func TestScripts(t *testing.T) {
 					ts.Fatalf("reading %q; got %q want %q", args[0], got, want)
 				}
 			},
-			"testscript-update": func(ts *TestScript, neg bool, args []string) {
+			"testscript": func(ts *TestScript, neg bool, args []string) {
 				// Run testscript in testscript. Oooh! Meta!
-				if len(args) != 1 {
-					ts.Fatalf("testscript <dir>")
+				fset := flag.NewFlagSet("testscript", flag.ContinueOnError)
+				fUpdate := fset.Bool("update", false, "update scripts when cmp fails")
+				fExplicitExec := fset.Bool("explicit-exec", false, "require explicit use of exec for commands")
+				fVerbose := fset.Bool("verbose", false, "be verbose with output")
+				if err := fset.Parse(args); err != nil {
+					ts.Fatalf("failed to parse args for testscript: %v", err)
 				}
-				t := &fakeT{ts: ts}
+				if fset.NArg() != 1 {
+					ts.Fatalf("testscript [-verbose] [-update] [-explicit-exec] <dir>")
+				}
+				dir := fset.Arg(0)
+				t := &fakeT{ts: ts, verbose: *fVerbose}
 				func() {
 					defer func() {
 						if err := recover(); err != nil {
@@ -190,18 +201,24 @@ func TestScripts(t *testing.T) {
 						}
 					}()
 					RunT(t, Params{
-						Dir:           ts.MkAbs(args[0]),
-						UpdateScripts: true,
+						Dir:                 ts.MkAbs(dir),
+						UpdateScripts:       *fUpdate,
+						RequireExplicitExec: *fExplicitExec,
+						Cmds: map[string]func(ts *TestScript, neg bool, args []string){
+							"some-param-cmd": func(ts *TestScript, neg bool, args []string) {
+							},
+						},
 					})
 				}()
+				ts.stdout = strings.Replace(t.log.String(), ts.workdir, "$WORK", -1)
 				if neg {
 					if len(t.failMsgs) == 0 {
-						ts.Fatalf("testscript-update unexpectedly succeeded")
+						ts.Fatalf("testscript unexpectedly succeeded")
 					}
 					return
 				}
 				if len(t.failMsgs) > 0 {
-					ts.Fatalf("testscript-update unexpectedly failed with errors: %q", t.failMsgs)
+					ts.Fatalf("testscript unexpectedly failed with errors: %q", t.failMsgs)
 				}
 			},
 		},
@@ -301,12 +318,30 @@ func TestBadDir(t *testing.T) {
 	if got := len(ft.failMsgs); got != wantCount {
 		t.Fatalf("expected %v fail message; got %v", wantCount, got)
 	}
-	wantMsg := regexp.MustCompile(`no scripts found matching glob: thiswillnevermatch[/\\]\*\.txt`)
+	wantMsg := regexp.MustCompile(`no txtar nor txt scripts found in dir thiswillnevermatch`)
 	if got := ft.failMsgs[0]; !wantMsg.MatchString(got) {
 		t.Fatalf("expected msg to match `%v`; got:\n%v", wantMsg, got)
 	}
 }
 
+func TestUNIX2DOS(t *testing.T) {
+	for data, want := range map[string]string{
+		"":         "",           // Preserve empty files.
+		"\n":       "\r\n",       // Convert LF to CRLF in a file containing a single empty line.
+		"\r\n":     "\r\n",       // Preserve CRLF in a single line file.
+		"a":        "a\r\n",      // Append CRLF to a single line file with no line terminator.
+		"a\n":      "a\r\n",      // Convert LF to CRLF in a file containing a single non-empty line.
+		"a\r\n":    "a\r\n",      // Preserve CRLF in a file containing a single non-empty line.
+		"a\nb\n":   "a\r\nb\r\n", // Convert LF to CRLF in multiline UNIX file.
+		"a\r\nb\n": "a\r\nb\r\n", // Convert LF to CRLF in a file containing a mix of UNIX and DOS lines.
+		"a\nb\r\n": "a\r\nb\r\n", // Convert LF to CRLF in a file containing a mix of UNIX and DOS lines.
+	} {
+		if got, err := unix2DOS([]byte(data)); err != nil || !bytes.Equal(got, []byte(want)) {
+			t.Errorf("unix2DOS(%q) == %q, %v, want %q, nil", data, got, err, want)
+		}
+	}
+}
+
 func setSpecialVal(ts *TestScript, neg bool, args []string) {
 	ts.Setenv("SPECIALVAL", "42")
 }
@@ -357,7 +392,10 @@ func waitFile(ts *TestScript, neg bool, args []string) {
 
 type fakeT struct {
 	ts       *TestScript
+	log      bytes.Buffer
 	failMsgs []string
+	verbose  bool
+	failed   bool
 }
 
 var errAbort = errors.New("abort test")
@@ -367,6 +405,7 @@ func (t *fakeT) Skip(args ...interface{}) {
 }
 
 func (t *fakeT) Fatal(args ...interface{}) {
+	t.failed = true
 	t.failMsgs = append(t.failMsgs, fmt.Sprint(args...))
 	panic(errAbort)
 }
@@ -374,7 +413,7 @@ func (t *fakeT) Fatal(args ...interface{}) {
 func (t *fakeT) Parallel() {}
 
 func (t *fakeT) Log(args ...interface{}) {
-	t.ts.Logf("testscript: %v", fmt.Sprint(args...))
+	fmt.Fprint(&t.log, args...)
 }
 
 func (t *fakeT) FailNow() {
@@ -386,5 +425,9 @@ func (t *fakeT) Run(name string, f func(T)) {
 }
 
 func (t *fakeT) Verbose() bool {
-	return false
+	return t.verbose
+}
+
+func (t *fakeT) Failed() bool {
+	return t.failed
 }
